import pytest
from types import SimpleNamespace
from imagetune.helper_functions import add_written_names, _param_list_excl_im, replace_in_args_or_kwargs, resolve_argname, find_in_args_or_kwargs


def f(im, a, b, c=5):
    return im * a + b * c

@pytest.fixture
def spec_good():
    # matches f(im, a, b, c=5)
    return SimpleNamespace(args=['im', 'a', 'b', 'c'])

@pytest.fixture
def spec_empty():
    # "wrong" argspec case you called out
    return SimpleNamespace(args=[])

@pytest.fixture
def spec_none():
    # also ensure robustness to args=None
    return SimpleNamespace(args=None)


# -------------------- _param_list_excl_im ----------------------

def test_param_list_excludes_first_param(spec_good):
    assert _param_list_excl_im(spec_good, None) == ['a', 'b', 'c']

def test_param_list_uses_placeholders_when_argnum_exceeds_length(spec_empty):
    assert _param_list_excl_im(spec_empty, 3) == ['arg1', 'arg2', 'arg3']

def test_param_list_with_argnum_but_sufficient_params_keeps_real_names(spec_good):
    assert _param_list_excl_im(spec_good, 2) == ['a', 'b', 'c']

def test_param_list_handles_none_args(spec_none):
    assert _param_list_excl_im(spec_none, 2) == ['arg1', 'arg2']


# ----------------------- resolve_argname -----------------------

def test_resolve_by_argnum(spec_good):
    assert resolve_argname(spec_good, argnum=2) == 'b'

def test_resolve_by_argname_direct(spec_good):
    assert resolve_argname(spec_good, argname='c') == 'c'

def test_resolve_with_placeholders_when_argspec_empty(spec_empty):
    assert resolve_argname(spec_empty, argnum=2) == 'arg2'

@pytest.mark.parametrize("argnum,argname", [(None, None), (1, 'a')])
def test_resolve_requires_exactly_one_selector(spec_good, argnum, argname):
    with pytest.raises(ValueError):
        resolve_argname(spec_good, argnum=argnum, argname=argname)


# ------------------- find_in_args_or_kwargs --------------------

def test_find_prefers_kwargs_when_present(spec_good):
    args = (2, 3, 7)
    kwargs = {'b': 42}
    assert find_in_args_or_kwargs(spec_good, args, kwargs, argname='b') == 42

def test_find_by_argnum_from_args(spec_good):
    args = (2, 3, 7)  # maps to a=2, b=3, c=7
    kwargs = {}
    assert find_in_args_or_kwargs(spec_good, args, kwargs, argnum=1) == 2

def test_find_by_argname_from_args(spec_good):
    args = (2, 3, 7)
    kwargs = {}
    assert find_in_args_or_kwargs(spec_good, args, kwargs, argname='b') == 3

def test_find_unknown_param_raises_keyerror(spec_good):
    with pytest.raises(KeyError):
        find_in_args_or_kwargs(spec_good, (), {}, argname='does_not_exist')

def test_find_value_missing_raises_keyerror(spec_good):
    # ask for c (idx 2) but provide fewer args and no kwargs
    with pytest.raises(KeyError):
        find_in_args_or_kwargs(spec_good, (1,), {}, argname='c')

def test_find_with_placeholder_params_when_argspec_empty(spec_empty):
    # argnum=2 -> params = ['arg1','arg2']; take from args position 1
    args = (10, 20)
    kwargs = {}
    assert find_in_args_or_kwargs(spec_empty, args, kwargs, argnum=2) == 20

def test_find_requires_exactly_one_selector_value_error(spec_good):
    with pytest.raises(ValueError):
        find_in_args_or_kwargs(spec_good, (), {}, argnum=None, argname=None)
    with pytest.raises(ValueError):
        find_in_args_or_kwargs(spec_good, (), {}, argnum=1, argname='a')


# ------------------ replace_in_args_or_kwargs ------------------

def test_replace_in_kwargs(spec_good):
    args = [2, 3, 7]
    kwargs = {'b': 9}
    replace_in_args_or_kwargs(spec_good, args, kwargs, 99, argname='b')
    assert kwargs['b'] == 99
    assert args == [2, 3, 7]

def test_replace_in_args_by_argnum(spec_good):
    args = [2, 3, 7]
    kwargs = {}
    replace_in_args_or_kwargs(spec_good, args, kwargs, 5, argnum=1)  # replace a
    assert args == [5, 3, 7]
    assert kwargs == {}

def test_replace_in_args_by_argname(spec_good):
    args = [2, 3, 7]
    kwargs = {}
    replace_in_args_or_kwargs(spec_good, args, kwargs, 11, argname='c')
    assert args == [2, 3, 11]

def test_replace_adds_to_kwargs_when_not_found_anywhere(spec_good):
    args = [2]  # only 'a' present positionally
    kwargs = {}
    replace_in_args_or_kwargs(spec_good, args, kwargs, 123, argname='c')
    assert kwargs == {'c': 123}
    assert args == [2]

def test_replace_with_placeholder_params_when_argspec_empty(spec_empty):
    # argnum=3 -> params ['arg1','arg2','arg3']; only one arg provided; should go to kwargs['arg3']
    args = [10]
    kwargs = {}
    replace_in_args_or_kwargs(spec_empty, args, kwargs, 777, argnum=3)
    assert args == [10]
    assert kwargs == {'arg3': 777}

def test_replace_requires_exactly_one_selector_value_error(spec_good):
    with pytest.raises(ValueError):
        replace_in_args_or_kwargs(spec_good, [], {}, 0, argnum=None, argname=None)
    with pytest.raises(ValueError):
        replace_in_args_or_kwargs(spec_good, [], {}, 0, argnum=1, argname='a')


# ----------------------- add_written_names ---------------------

def test_add_written_names_deduplicates_with_suffixes():
    d = {
        'x': {'name': 'alpha'},
        'y': {'name': 'beta'},
        'z': {'name': 'alpha'},
        'w': {'name': 'beta'},
        'u': {'name': 'gamma'},
    }
    out = add_written_names(d)

    # counts: alpha=2, beta=2 -> suffixed; gamma=1 -> unsuffixed
    assert out['x']['written_name'] == 'alpha:1'
    assert out['z']['written_name'] == 'alpha:2'
    assert out['y']['written_name'] == 'beta:1'
    assert out['w']['written_name'] == 'beta:2'
    assert out['u']['written_name'] == 'gamma'

def test_add_written_names_returns_same_object_reference():
    d = {'k': {'name': 'n'}}
    out = add_written_names(d)
    assert out is d  # function mutates and returns the same dict



from PyQt6 import QtCore
import pytest
from skimage import data

from imagetune.imagetune import _tune_ui_widget
from imagetune import tune


def adjust(im, alpha, gamma):
    return alpha * im**gamma

@tune(argnums=(2, ))
def adjust2(im, alpha, gamma):
    return alpha * im**gamma

@tune
def adjust3(im, alpha, gamma):
    return alpha * im**gamma

@tune
def adjust4(im, alpha, gamma=2.0):
    return alpha * im**gamma


def preprocess(im):
    im = tune(adjust)(im, alpha=0.5, gamma=1.0)
    im = tune(adjust, argnames='gamma')(im, alpha=0.5, gamma=1.0)
    im = tune(adjust, argnums=(1, 2))(im, alpha=0.5, gamma=1.0)
    im = tune(adjust, argnames='gamma')(im, 0.5, 1.0)
    im = tune(adjust, argnums=(1, 2))(im, 0.5, gamma=1.0)
    im = adjust2(im, 0.5, gamma=1.0)
    im = adjust3(im, 0.5, 1.0)
    im = adjust4(im, 0.5)
    return im


@pytest.mark.parametrize("show_window", [True, False])
def test_main_widget_starts_and_closes(qtbot, show_window):
    im = data.astronaut()[:, :, 0] / 255.0
    w = _tune_ui_widget(preprocess, im)

    qtbot.addWidget(w)

    if show_window:
        w.show()
        try:
            qtbot.waitExposed(w, timeout=3000)
        except Exception:
            pass  # okay on offscreen

    qtbot.wait(100)

    w.close()



import numpy as np 
from functools import wraps
from collections import OrderedDict
import inspect

from .ui import make_ui, _build_ui_widget

_TUNES = OrderedDict()
_INFO = {'INDEX' : 0}


def tune(func=None, min=None, max=None, name=None, argnums=None, argnames=None):
    if argnames is not None:
        assert argnums is None, "Only one of (argnums, argnames) can be specified."

    if argnums is None and argnames is None:
        argnums = (1, )

    if argnums is not None:
        if isinstance(argnums, int):
            argnums = (argnums,)

        assert isinstance(argnums, tuple), "`argnums` must be a tuple of integers"
        assert isinstance(argnums[0], int), "`argnums` must be a tuple of integers"
        assert all(x >= 1 for x in argnums), "`argnums` must be larger than zero (first argument should be image)"

    if argnames is not None:
        if isinstance(argnames, str):
            argnames = (argnames,)

        assert isinstance(argnames, tuple), "`argnames` must be a tuple of integers"
        assert isinstance(argnames[0], str), "`argnames` must be a tuple of strings"


    def decorator(f):
        @wraps(f)
        def wrapper(im, *args, **kwargs):
            idx = _INFO["INDEX"]
            name_ = name or f.__name__

            if (name_, idx) not in _TUNES:
                # Fill on first call (ensures order is correct):
                _TUNES[(name_, idx)] = {"name": name_, "func": wrapper, "min": min, "max": max,
                                        "args": list(args), "kwargs": dict(kwargs), "result": None, "index": idx,
                                        "argnums":argnums, "argnames": argnames, "argspec": inspect.getfullargspec(f)}

            res = f(im, *_TUNES[(name_, idx)]['args'], **_TUNES[(name_, idx)]['kwargs'])
            _TUNES[(name_, idx)]['result'] = np.array(res)
            _INFO['INDEX'] += 1

            return res

        return wrapper

    if func is None:
        # Case: @tune(min=..., max=...)
        return decorator
    elif callable(func):
        # Case: tune(func, min=..., max=...)
        return decorator(func)
    else:
        raise TypeError("Incorrect use of `tune` decorator.")


def tuneui(pipeline, im):
    def wrappred(im):
        _INFO['INDEX'] = 0
        return pipeline(im)

    wrappred(im)
    make_ui(wrappred, im, _TUNES)


def _tune_ui_widget(pipeline, im):
    def wrappred(im):
        _INFO['INDEX'] = 0
        return pipeline(im)

    wrappred(im)
    return _build_ui_widget(wrappred, im, _TUNES)




from PyQt6 import QtCore
import pytest
from skimage import data
from skimage.filters import gaussian, unsharp_mask

from imagetune.imagetune import _tune_ui_widget
from imagetune import tune


@tune(min=0, max=1.0)
def threshold(im, t1):
    return im > t1

@tune(argnums=(1, 2))
def adjust(im, alpha, gamma):
    return alpha * im**gamma


def preprocess(im):
    im = adjust(im, 1.0, 1.0)
    im = tune(gaussian)(im, 1.0)
    im = tune(unsharp_mask, argnames='amount')(im, radius=2.0, amount=1.0)
    im = threshold(im, 0.5)
    return im

@pytest.mark.parametrize("show_window", [True, False])
def test_main_widget_starts_and_closes(qtbot, show_window):
    im = data.astronaut()[:, :, 0] / 255.0
    w = _tune_ui_widget(preprocess, im)

    qtbot.addWidget(w)

    if show_window:
        w.show()
        try:
            qtbot.waitExposed(w, timeout=3000)
        except Exception:
            pass  # okay on offscreen

    qtbot.wait(100)

    w.close()



import numpy as np
import fastplotlib as fpl
from PySide6 import QtWidgets, QtCore
from functools import partial
from skimage.util import img_as_float

from .helper_functions import find_in_args_or_kwargs, replace_in_args_or_kwargs, resolve_argname, add_written_names


def _from_slider(val, min, max):
    return min + val / 1000.0 * (max - min)


def _to_slider(val, min, max):
    return int(1000.0 * (val - min) / (max - min))


def _build_ui_widget(pipeline, im, tunes):
    im = img_as_float(im)

    add_written_names(tunes)

    intermediate_plot = True  # for now always do intermediate

    fig = fpl.Figure(shape=(1, 3 if intermediate_plot else 2), controller_ids="sync")
    if intermediate_plot:
        ax_orig, ax_intermediate, ax_final = fig[0, 0], fig[0, 1], fig[0, 2]
        ax_intermediate.title = "intermediate"
        bin_intermediate = [ax_intermediate.add_image(im)]
    else:
        ax_orig, ax_intermediate, ax_final = fig[0, 0], None, fig[0, 1]

    for ax in (ax_orig, ax_intermediate, ax_final):
        if ax is not None:
            ax.axes.visible = False

    bin_orig = ax_orig.add_image(im)
    bin_final = [ax_final.add_image(im)]
    ax_orig.title = "original"
    ax_final.title = "final"
    canvas = fig.show()
    w = QtWidgets.QWidget()
    lay = QtWidgets.QVBoxLayout(w)
    lay.addWidget(canvas)

    def update_image(tune):
        r_im = pipeline(im)
        try:
            bin_final[0].data = img_as_float(r_im)
        except ValueError:
            ax_final.clear()
            bin_final[0] = ax_final.add_image(img_as_float(r_im))

        if intermediate_plot:
            try:
                bin_intermediate[0].data = img_as_float(tune['result'])
            except ValueError:
                ax_intermediate.clear()
                bin_intermediate[0] = ax_intermediate.add_image(img_as_float(tune['result']))

            ax_intermediate.title = f"{tune['index'] + 1} : {tune['written_name']}"

    def update(v, tune, label, arg_index):
        use_argnames = tune['argnames'] is not None
        argnum = tune['argnums'][arg_index] if not use_argnames else None
        argname = tune['argnames'][arg_index] if use_argnames else None

        v = _from_slider(v, tune['min'], tune['max'])
        replace_in_args_or_kwargs(tune['argspec'], tune['args'], tune['kwargs'], v,
                                  argnum=argnum, argname=argname)

        argname = resolve_argname(tune['argspec'], argnum=argnum, argname=argname)
        label.setText(f"{tune['index'] + 1} : {tune['written_name']} ({argname}): {v:.3f}")

        update_image(tune)

    for tune in tunes.values():
        use_argnames = tune['argnames'] is not None
        n_tunable = len(tune['argnames']) if use_argnames else len(tune['argnums'])

        for arg_index in range(n_tunable):
            layout = QtWidgets.QHBoxLayout()
            label = QtWidgets.QLabel("")
            slider = QtWidgets.QSlider(QtCore.Qt.Orientation.Horizontal)

            if use_argnames:
                value = find_in_args_or_kwargs(tune['argspec'], tune['args'], tune['kwargs'],
                                               argname=tune['argnames'][arg_index])
            else:
                value = find_in_args_or_kwargs(tune['argspec'], tune['args'], tune['kwargs'],
                                               argnum=tune['argnums'][arg_index])

            if tune['min'] is None:
                tune['min'] = 0.1 * value

            if tune['max'] is None:
                tune['max'] = 1 if abs(value) < 1e-5 else 10 * value

            slider.setRange(0, 1000)
            slider.setValue(_to_slider(value, tune['min'], tune['max']))
            layout.addWidget(label)
            layout.addWidget(slider)
            lay.addLayout(layout)

            slider.valueChanged.connect(partial(update, tune=tune, label=label, arg_index=arg_index))
            v = _to_slider(value, tune['min'], tune['max'])
            slider.sliderPressed.connect(partial(update, v=v, tune=tune, label=label, arg_index=arg_index))
            update(v=v, tune=tune, label=label, arg_index=arg_index)

    return w

def make_ui(pipeline, im, tunes, width=1200, height=500):
    app = QtWidgets.QApplication([])
    w = _build_ui_widget(pipeline, im, tunes)
    w.setWindowTitle("ImageTune")
    w.resize(width, height)
    w.show()
    app.exec()




def _param_list_excl_im(argspec, argnum):
    a = list(argspec.args or [])
    if a:
        a = a[1:]

    if argnum is not None and not (len(a) >= argnum):
        a = [f'arg{i}' for i in range(1, argnum + 1)]

    return a


def resolve_argname(argspec, argnum=None, argname=None):
    params = _param_list_excl_im(argspec, argnum)
    if (argnum is None) == (argname is None):
        raise ValueError("Provide exactly one of argnum or argname")
    return argname if argname is not None else params[argnum - 1]


def find_in_args_or_kwargs(argspec, args, kwargs, argnum=None, argname=None):
    params = _param_list_excl_im(argspec, argnum)

    if (argnum is None) == (argname is None):
        raise ValueError("Provide exactly one of argnum or argname")

    name = argname if argname is not None else params[argnum - 1]
    if name in kwargs:
        return kwargs[name]

    try:
        idx = params.index(name) if argname is not None else argnum - 1
    except ValueError:
        raise KeyError("Unknown parameter: %r" % name)

    if 0 <= idx < len(args):
        return args[idx]

    raise KeyError("Value for %r not found in args or kwargs" % name)


def replace_in_args_or_kwargs(argspec, args, kwargs, new_value, argnum=None, argname=None):
    params = _param_list_excl_im(argspec, argnum)
    if (argnum is None) == (argname is None):
        raise ValueError("Provide exactly one of argnum or argname")

    name = argname if argname is not None else params[argnum - 1]

    if name in kwargs:
        kwargs[name] = new_value
        return

    idx = params.index(name) if argname is not None else argnum - 1
    if 0 <= idx < len(args):
        args[idx] = new_value
        return

    kwargs[name] = new_value


def add_written_names(d):
    counts = {}

    for v in d.values():
        counts[v["name"]] = counts.get(v["name"], 0) + 1
    seen = {}

    for v in d.values():
        n = v["name"]
        seen[n] = seen.get(n, 0) + 1
        v["written_name"] = f"{n}:{seen[n]}" if counts[n] > 1 else n

    return d
